# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ava_label.proto
# Protobuf Python Version: 4.25.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0f\x61va_label.proto"\xa0\x01\n\x05Label\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08label_id\x18\x02 \x01(\r\x12$\n\nlabel_type\x18\x03 \x01(\x0e\x32\x10.Label.LabelType"Q\n\tLabelType\x12\x13\n\x0fPERSON_MOVEMENT\x10\x00\x12\x17\n\x13OBJECT_MANIPULATION\x10\x01\x12\x16\n\x12PERSON_INTERACTION\x10\x02""\n\tLabelList\x12\x15\n\x05label\x18\x01 \x03(\x0b\x32\x06.Labelb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "ava_label_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS is False:
    DESCRIPTOR._options = None
    _globals["_LABEL"]._serialized_start = 20
    _globals["_LABEL"]._serialized_end = 180
    _globals["_LABEL_LABELTYPE"]._serialized_start = 99
    _globals["_LABEL_LABELTYPE"]._serialized_end = 180
    _globals["_LABELLIST"]._serialized_start = 182
    _globals["_LABELLIST"]._serialized_end = 216
# @@protoc_insertion_point(module_scope)
